/*
 * Copyright (C) 2020 Tycho Softworks.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef MEMORY_HPP_
#define MEMORY_HPP_

#include <memory>
#include <string>

#include "server.hpp"

namespace server {
class mempager {
public:
    mempager(size_t size = 0) noexcept;
    mempager(mempager&& move) noexcept;
    virtual ~mempager();

    mempager(const mempager& copy) noexcept = delete;
    auto operator=(const mempager& copy) -> mempager& = delete;
    auto operator=(mempager&& move) noexcept -> mempager&;

    inline auto used() const {
        return page_count * page_size;
    }

    auto alloc(size_t size) -> void *;
    auto dup(const char *str) -> char *;
    auto dup(const std::string& str) -> char *;
    void clear();

    template<typename T>
    inline auto make(size_t adjust) {
        return static_cast<T*>(alloc(sizeof(T) + adjust));
    }

    inline static auto aligned_size(size_t size) {
        size_t align = 1;
        while(align < size)
            align <<= 1;
        return align;
    }

    inline static auto aligned_page(size_t min = 0) {
#if defined(_SC_PAGESIZE)
        size_t size = aligned_size(static_cast<size_t>(sysconf(_SC_PAGESIZE)));
#elif defined(PAGESIZE)
        size_t size = aligned_size(PAGESIZE);
#elif defined(PAGE_SIZE)
        size_t size = aligned_size(PAGE_SIZE);
#else
        size_t size = 1024;
#endif
        min = aligned_size(min);
        while(size < min)
            size <<= 2;
        return size;
    }

    inline static auto aligned_cache() {
        static volatile size_t line_size = 0;
#if defined(_SC_LEVEL1_DCHACHE_LINESIZE)
        line_size = sysconf(_SC_LEVEL1_DCACHE_LINESIZE);
#endif
        if(line_size < 64)
            line_size = 64;
        return aligned_size(line_size);
    }

    static auto aligned_alloc(size_t align, size_t size) -> void *;

protected:
    size_t page_size, mem_align;
    unsigned page_count;

private:
    using page_ptr = struct page_t {
        page_t *next;
        union {
            void *aligned;
            unsigned used;
        };
    } *;

    page_ptr current_;
};

auto mem_index(const uint8_t *mem, size_t size) -> unsigned;
auto mem_size(const char *cp, size_t max = 256) -> size_t;
auto mem_append(char *dest, size_t max, const char *value) -> size_t;
auto mem_value(char *target, size_t size, unsigned value) -> bool;
auto mem_copy(char *target, size_t size, const char *from) -> bool;
auto mem_view(char *target, size_t size, std::string_view s) -> bool;

inline static auto mem_index(std::string_view s) {
    return mem_index(reinterpret_cast<const uint8_t *>(s.data()), s.size());
}

inline static auto key_index(const char *cp, size_t max) {
    return mem_index(reinterpret_cast<const uint8_t *>(cp), mem_size(cp, max));
}
}

inline auto operator new(size_t size, server::mempager& pager) -> void * {
    return pager.alloc(size);
}

inline void operator delete([[maybe_unused]] void *page, [[maybe_unused]] server::mempager& pager) {}
#endif
